@@ -34,25 +34,50 @@ struct ParticleLightData
     float spotLightInnerConeAngle[4] = {0.0f};
 };
 
-void QSSGParticleRenderer::updateUniformsForParticles(QSSGRhiShaderPipeline &shaders,
+void QSSGParticleRenderer::updateUniformsForParticles(const QSSGLayerRenderData &inData,
+                                                      QSSGRhiShaderPipeline &shaders,
                                                       QSSGRhiContext *rhiCtx,
                                                       char *ubufData,
                                                       QSSGParticlesRenderable &renderable,
-                                                      QSSGRenderCamera &inCamera)
+                                                      const QSSGRenderCameraList &cameras)
 {
     const QMatrix4x4 clipSpaceCorrMatrix = rhiCtx->rhi()->clipSpaceCorrMatrix();
 
     QSSGRhiShaderPipeline::CommonUniformIndices &cui = shaders.commonUniformIndices;
 
-    const QMatrix4x4 projection = clipSpaceCorrMatrix * inCamera.projection;
-    shaders.setUniform(ubufData, "qt_projectionMatrix", projection.constData(), 16 * sizeof(float), &cui.projectionMatrixIdx);
+    const int viewCount = cameras.count();
 
-    const QMatrix4x4 viewMatrix = inCamera.globalTransform.inverted();
-    shaders.setUniform(ubufData, "qt_viewMatrix", viewMatrix.constData(), 16 * sizeof(float), &cui.viewMatrixIdx);
+    // Pull the camera transforms from the render data once.
+    QMatrix4x4 camGlobalTransforms[2] { QMatrix4x4{Qt::Uninitialized}, QMatrix4x4{Qt::Uninitialized} };
+    if (viewCount < 2) {
+        camGlobalTransforms[0] = inData.getGlobalTransform(*cameras[0]);
+    } else {
+        for (size_t viewIndex = 0; viewIndex != std::size(camGlobalTransforms); ++viewIndex)
+            camGlobalTransforms[viewIndex] = inData.getGlobalTransform(*cameras[viewIndex]);
+    }
+
+    if (viewCount < 2) {
+        const QMatrix4x4 projection = clipSpaceCorrMatrix * cameras[0]->projection;
+        shaders.setUniform(ubufData, "qt_projectionMatrix", projection.constData(), 16 * sizeof(float), &cui.projectionMatrixIdx);
+        const QMatrix4x4 viewMatrix = camGlobalTransforms[0].inverted();
+        shaders.setUniform(ubufData, "qt_viewMatrix", viewMatrix.constData(), 16 * sizeof(float), &cui.viewMatrixIdx);
+    } else {
+        QVarLengthArray<QMatrix4x4, 2> projectionMatrices(viewCount);
+        QVarLengthArray<QMatrix4x4, 2> viewMatrices(viewCount);
+        for (int viewIndex = 0; viewIndex < viewCount; ++viewIndex) {
+            projectionMatrices[viewIndex] = clipSpaceCorrMatrix * cameras[viewIndex]->projection;
+            viewMatrices[viewIndex] = camGlobalTransforms[viewIndex].inverted();
+        }
+        shaders.setUniformArray(ubufData, "qt_projectionMatrix", projectionMatrices.constData(), viewCount, QSSGRenderShaderValue::Matrix4x4, &cui.projectionMatrixIdx);
+        shaders.setUniformArray(ubufData, "qt_viewMatrix", viewMatrices.constData(), viewCount, QSSGRenderShaderValue::Matrix4x4, &cui.viewMatrixIdx);
+    }
 
     const QMatrix4x4 &modelMatrix = renderable.globalTransform;
     shaders.setUniform(ubufData, "qt_modelMatrix", modelMatrix.constData(), 16 * sizeof(float), &cui.modelMatrixIdx);
 
+    const QVector2D camProperties(cameras[0]->clipPlanes);
+    shaders.setUniform(ubufData, "qt_cameraProperties", &camProperties, 2 * sizeof(float), &cui.cameraPropertiesIdx);
+
     QVector2D oneOverSize = QVector2D(1.0f, 1.0f);
     auto &particleBuffer = renderable.particles.m_particleBuffer;
     const quint32 particlesPerSlice = particleBuffer.particlesPerSlice();
@@ -87,18 +112,21 @@ void QSSGParticleRenderer::updateUniformsForParticles(QSSGRhiShaderPipeline &sha
             // Ignore lights which are not specified for the particle
             if (!renderable.particles.m_lights.contains(theLight))
                 continue;
+
+            const QMatrix4x4 lightGlobalTransform = inData.getGlobalTransform(*theLight);
+
             if (theLight->type == QSSGRenderLight::Type::DirectionalLight) {
                 theLightAmbientTotal += theLight->m_diffuseColor * theLight->m_brightness;
             } else if (theLight->type == QSSGRenderLight::Type::PointLight && pointLight < 4) {
                 lightData.pointLightColor[pointLight] = QVector4D(theLight->m_diffuseColor * theLight->m_brightness, 1.0f);
-                lightData.pointLightPos[pointLight] = QVector4D(theLight->getGlobalPos(), 1.0f);
+                lightData.pointLightPos[pointLight] = QVector4D(QSSGRenderNode::getGlobalPos(lightGlobalTransform), 1.0f);
                 lightData.pointLightConstantAtt[pointLight] = QSSGUtils::aux::translateConstantAttenuation(theLight->m_constantFade);
                 lightData.pointLightLinearAtt[pointLight] = QSSGUtils::aux::translateLinearAttenuation(theLight->m_linearFade);
                 lightData.pointLightQuadAtt[pointLight] = QSSGUtils::aux::translateQuadraticAttenuation(theLight->m_quadraticFade);
                 pointLight++;
             } else if (theLight->type == QSSGRenderLight::Type::SpotLight && spotLight < 4) {
                 lightData.spotLightColor[spotLight] = QVector4D(theLight->m_diffuseColor * theLight->m_brightness, 1.0f);
-                lightData.spotLightPos[spotLight] = QVector4D(theLight->getGlobalPos(), 1.0f);
+                lightData.spotLightPos[spotLight] = QVector4D(QSSGRenderNode::getGlobalPos(lightGlobalTransform), 1.0f);
                 lightData.spotLightDir[spotLight] = QVector4D(lights[lightIdx].direction, 0.0f);
                 lightData.spotLightConstantAtt[spotLight] = QSSGUtils::aux::translateConstantAttenuation(theLight->m_constantFade);
                 lightData.spotLightLinearAtt[spotLight] = QSSGUtils::aux::translateLinearAttenuation(theLight->m_linearFade);
@@ -182,11 +210,12 @@ static void fillTargetBlend(QRhiGraphicsPipeline::TargetBlend &targetBlend, QSSG
     }
 }
 
-static void sortParticles(QByteArray &result, QList<QSSGRhiSortData> &sortData,
+static void sortParticles(const QSSGLayerRenderData &renderData, QByteArray &result, QList<QSSGRhiSortData> &sortData,
                           const QSSGParticleBuffer &buffer, const QSSGRenderParticles &particles,
                           const QVector3D &cameraDirection, bool animatedParticles)
 {
-    const QMatrix4x4 &invModelMatrix = particles.globalTransform.inverted();
+    const QMatrix4x4 &modelMatrix = renderData.getGlobalTransform(particles);
+    const QMatrix4x4 &invModelMatrix = modelMatrix.inverted();
     QVector3D dir = invModelMatrix.map(cameraDirection);
     QVector3D n = dir.normalized();
     const auto segments = buffer.segments();
@@ -329,7 +358,8 @@ void QSSGParticleRenderer::rhiPrepareRenderable(QSSGRhiShaderPipeline &shaderPip
                                                 const QSSGLayerRenderData &inData,
                                                 QRhiRenderPassDescriptor *renderPassDescriptor,
                                                 int samples,
-                                                QSSGRenderCamera *camera,
+                                                int viewCount,
+                                                QSSGRenderCamera *alteredCamera,
                                                 QSSGRenderTextureCubeFace cubeFace,
                                                 QSSGReflectionMapEntry *entry)
 {
@@ -341,10 +371,12 @@ void QSSGParticleRenderer::rhiPrepareRenderable(QSSGRhiShaderPipeline &shaderPip
     shaderPipeline.ensureUniformBuffer(&dcd.ubuf);
 
     char *ubufData = dcd.ubuf->beginFullDynamicBufferUpdateForCurrentFrame();
-    if (!camera)
-        updateUniformsForParticles(shaderPipeline, rhiCtx, ubufData, renderable, *inData.camera);
-    else
-        updateUniformsForParticles(shaderPipeline, rhiCtx, ubufData, renderable, *camera);
+    if (!alteredCamera) {
+        updateUniformsForParticles(inData, shaderPipeline, rhiCtx, ubufData, renderable, inData.renderedCameras);
+    } else {
+        QSSGRenderCameraList cameras({ alteredCamera });
+        updateUniformsForParticles(inData, shaderPipeline, rhiCtx, ubufData, renderable, cameras);
+    }
     dcd.ubuf->endFullDynamicBufferUpdateForCurrentFrame();
 
     QSSGRhiParticleData &particleData = QSSGRhiContextPrivate::get(rhiCtx)->particleData(&renderable.particles);
@@ -373,10 +405,12 @@ void QSSGParticleRenderer::rhiPrepareRenderable(QSSGRhiShaderPipeline &shaderPip
 
     if (renderable.particles.m_depthSorting) {
         bool animatedParticles = renderable.particles.m_featureLevel == QSSGRenderParticles::FeatureLevel::Animated;
-        if (!camera)
-            sortParticles(particleData.sortedData, particleData.sortData, particleBuffer, renderable.particles, inData.cameraData->direction, animatedParticles);
-        else
-            sortParticles(particleData.sortedData, particleData.sortData, particleBuffer, renderable.particles, camera->getScalingCorrectDirection(), animatedParticles);
+        if (!alteredCamera) {
+            sortParticles(inData, particleData.sortedData, particleData.sortData, particleBuffer, renderable.particles, inData.renderedCameraData.value()[0].direction, animatedParticles);
+        } else {
+            const QMatrix4x4 globalTransform = inData.getGlobalTransform(*alteredCamera);
+            sortParticles(inData, particleData.sortedData, particleData.sortData, particleBuffer, renderable.particles, QSSGRenderNode::getScalingCorrectDirection(globalTransform), animatedParticles);
+        }
         uploadData = convertParticleData(particleData.convertData, particleData.sortedData, needsConversion);
     } else {
         uploadData = convertParticleData(particleData.convertData, particleBuffer.data(), needsConversion);
@@ -395,11 +429,15 @@ void QSSGParticleRenderer::rhiPrepareRenderable(QSSGRhiShaderPipeline &shaderPip
     ia.inputs.clear();
 
     ps->samples = samples;
+    ps->viewCount = viewCount;
     ps->cullMode = QRhiGraphicsPipeline::None;
-    if (renderable.renderableFlags.hasTransparency())
-        fillTargetBlend(ps->targetBlend, renderable.particles.m_blendMode);
-    else
-        ps->targetBlend = QRhiGraphicsPipeline::TargetBlend();
+
+    if (inData.orderIndependentTransparencyEnabled == false) {
+        if (renderable.renderableFlags.hasTransparency())
+            fillTargetBlend(ps->targetBlend[0], renderable.particles.m_blendMode);
+        else
+            ps->targetBlend[0] = QRhiGraphicsPipeline::TargetBlend();
+    }
 
     QSSGRhiShaderResourceBindingList bindings;
     bindings.addUniformBuffer(0, VISIBILITY_ALL, dcd.ubuf, 0, shaderPipeline.ub0Size());
@@ -491,6 +529,7 @@ void QSSGParticleRenderer::rhiPrepareRenderable(QSSGRhiShaderPipeline &shaderPip
     bool srbChanged = false;
     if (!srb || bindings != dcd.bindings) {
         srb = rhiCtxD->srb(bindings);
+        rhiCtxD->releaseCachedSrb(dcd.bindings);
         dcd.bindings = bindings;
         srbChanged = true;
     }
